# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from typing import Any
from typing import Generator

import aorta
import fastapi
import fastapi.params
from fastapi.dependencies.models import Dependant
from fastapi.dependencies.utils import get_dependant
from fastapi.dependencies.utils import get_parameterless_sub_dependant
from fastapi.dependencies.utils import solve_dependencies

from cbra.types import IDependant
from .ioc.instance import Instance


class MessageRunner(aorta.LocalRunner, IDependant):
    __module__: str = 'cbra.ext.google'
    request: fastapi.Request

    def __init__(
        self,
        request: fastapi.Request,
        publisher: aorta.types.IPublisher = Instance.get('MessagePublisher')
    ):
        super().__init__(publisher=publisher)
        self.request = request

    async def handle(
        self,
        transaction: aorta.types.ITransaction,
        handler: aorta.types.IMessageHandler,
        envelope: aorta.types.Envelope[Any]
    ) -> tuple[bool, Any]:
            async with transaction:
                dependant = get_dependant(call=handler.run, path='/')
                dependant.dependencies.extend(self.get_injectors(handler))
                values, errors, *_ = await solve_dependencies(
                    request=self.request,
                    dependant=dependant,
                    # TODO: This assumes very specific internal FastAPI behaviors.
                    body=envelope, # type: ignore
                    dependency_overrides_provider=None
                )
                for error in errors:
                    self.logger.exception(
                        "Caught fatal %s while injecting dependencies in "
                        "%s: %s",
                        type(error.exc).__name__, type(handler).__name__,
                        repr(error.exc)
                    )
                    raise error.exc
                assert dependant.call is not None
                assert callable(dependant.call)
                return await dependant.call(**values)

    def get_injectors(self, obj: Any) -> Generator[Dependant, None, None]:
        for attname, member in inspect.getmembers(obj):
            if not isinstance(member, fastapi.params.Depends):
                continue
            def setter(attname: str, dep: Any = member):
                async def f(dep: Any = dep) -> None:
                    setattr(obj, attname, dep)
                return f
            yield get_parameterless_sub_dependant(
                 depends=fastapi.Depends(setter(attname, member)),
                 path='/'
            )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from types import NotImplementedType

import pydantic

from .basemodel import BaseModel
from .modelidentity import ModelIdentity
from .modelautoassignedidentity import ModelAutoAssignedIdentity
from .modelmetadata import ModelMetadata


class PersistedModel(BaseModel):
    __abstract__: bool = True
    __surrogate__: ModelAutoAssignedIdentity | NotImplementedType | None | int
    __key__: ModelIdentity | NotImplementedType
    __metadata__: ModelMetadata = pydantic.PrivateAttr(
        default_factory=ModelMetadata.initialize
    )

    def natural_key(self) -> tuple[str, ...]:
        raise NotImplementedError


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# type: ignore
import importlib
import inspect
import pathlib
import os
from typing import cast
from typing import Any
from typing import Callable
from typing import TypeVar

import aorta
import unimatrix.runtime
import yaml
from pydantic import Field

from cbra.types import Request
from .router import APIRouter
from .application import Application
from .endpoint import Endpoint
from .endpointtype import EndpointType
from . import ioc
from . import utils
from .localmessagetransport import LocalMessageTransport
from .messagepublisher import MessagePublisher
from .messagerunner import MessageRunner
from .metricreporter import MetricReporter
from .params import *
from .resource import Collection
from .resource import Create
from .resource import Delete
from .resource import Mutable
from .resource import QueryResult
from .resource import Replace
from .resource import Resource
from .resource import ResourceModel
from .resource import ResourceType
from .resource import Retrieve
from .resource import Update
from .resource import Versioned
from .secretkey import SecretKey


T = TypeVar('T')


__all__: list[str] = [
    'describe',
    'inject',
    'instance',
    'ioc',
    'permission',
    'run',
    'utils',
    'APIRouter',
    'Application',
    'ApplicationSecretKey',
    'Collection',
    'Create',
    'Delete',
    'Endpoint',
    'EndpointType',
    'Field',
    'LocalMessageTransport',
    'MessagePublisher',
    'MessageRunner',
    'MetricReporter',
    'Mutable',
    'QueryResult',
    'Request',
    'Replace',
    'Resource',
    'ResourceModel',
    'ResourceType',
    'Retrieve',
    'SecretKey',
    'Update',
    'Versioned',
]

inject = ioc.inject
instance = ioc.instance
permission = Endpoint.require_permission


class describe:
    status_code: int = 200
    summary: str | None = None

    def __init__(
        self,
        status_code: int = 200,
        summary: str | None = None
    ) -> None:
        self.status_code = status_code

    def __call__(
        self,
        func: Callable[..., T]
    ) -> Callable[..., T]:
        if not hasattr(func, 'params'):
            func.params = {} # type: ignore
        func.params.update({ # type: ignore
            k: v for k, v in self.__dict__.items()
            if v is not None
        })
        return func


class action:
    """Like :class:`describe`, but decorates the method to indicate that
    it exposes an action on a resource.
    """
    action: str
    method: str
    path: str
    status_code: int = 200
    summary: str | None = None

    def __init__(
        self,
        summary: str,
        *,
        method: str,
        path: str | None = None,
        status_code: int = 200,
    ) -> None:
        self.method = method
        self.path = path
        self.status_code = status_code
        self.summary = summary

    def __call__(
        self,
        func: Callable[..., T]
    ) -> Callable[..., T]:
        self.action = self.path or func.__name__
        func.action = {} # type: ignore
        func.action.update({ # type: ignore
            k: v for k, v in self.__dict__.items()
            if v is not None and k not in {'path'}
        })
        return func
    

class response:
    status_code: int
    description: str

    def __init__(
        self,
        status_code: int,
        description: str
    ) -> None:
        self.description = description
        self.status_code = status_code

    def __call__(
        self,
        func: Callable[..., T]
    ) -> Callable[..., T]:
        if not hasattr(func, 'responses'):
            func.responses = {} # type: ignore
        func.responses[self.status_code] = { # type: ignore
            k: v for k, v in self.__dict__.items()
            if v is not None and k != 'status_code'
        }
        return func
    

def autodiscover(
    qualname: str, cls: type[T] | None  = None
) -> T:
    os.environ.setdefault(
        'PYTHON_SETTINGS_MODULE',
        unimatrix.runtime.get_settings_module(__name__)
    )
    parent = qualname.rsplit('.', 1)[0]
    modules: list[str] = [
        'endpoints',
        'handlers',
        'listeners',
        'resources',
    ]
    asgi = importlib.import_module(qualname)
    params: dict[str, Any] = {
        'docs_url'  : '/ui',
        'redoc_url' : '/'
    }
    base = pathlib.Path(asgi.__file__).parent
    meta = base.joinpath('asgi.yml')
    if meta.exists():
        params.update(yaml.safe_load(meta.read_text()))
    cls = cls or Application
    app = cast(Application, cls(**params))
    addables: tuple[type] = (
        Resource,
        aorta.CommandHandler,
        aorta.Sewer,
        aorta.EventListener,
    )
    for name in modules:
        try:
            module = importlib.import_module(f'{parent}.{name}')
        except ImportError:
            # Raise the error if the module exists because then it
            # was something in the module that raised it.
            if any([
                base.joinpath(name).exists(),
                base.joinpath(f'{name}.py').exists()
            ]):
                raise
            continue
        for _, member in inspect.getmembers(module):
            if not inspect.isclass(member) or not issubclass(member, addables):
                continue
            if not getattr(member, 'autodiscover', False):
                continue
            app.add(member)

    return app


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any
from typing import Literal

import pydantic
from canonical import DomainName
from ckms.types import JSONWebKeySet
from headless.ext.oauth2.types import GrantType

from cbra.core.iam.models import AccessControlList
from ..types import ClientAuthenticationMethod
from ..types import FatalAuthorizationException
from ..types import PairwiseIdentifier
from ..types import RedirectURI
from ..types import RefreshTokenPolicyType
from ..types import RequestedScope
from .baseclient import BaseClient
from .downstreamproviderconfig import DownstreamProviderConfig


class ApplicationClient(BaseClient):
    """An OAuth 2.0 that is a specific application."""
    acl: AccessControlList = pydantic.Field(
        default_factory=lambda: AccessControlList(enabled=False)
    )
    allowed_resources: list[str] | None = None
    allowed_redirect_uris: list[RedirectURI] = pydantic.Field(
        default=[],
        min_length=1
    )
    client_id: str
    client_name: str
    client_secret: str | dict[str, Any] | None = None
    consent_required: bool = True
    contacts: list[str] = []
    downstream: DownstreamProviderConfig = pydantic.Field(
        default_factory=DownstreamProviderConfig
    )
    grant_types: list[GrantType] = []
    jwks: JSONWebKeySet | None = None
    kind: Literal['Application']
    organization_name: str
    refresh_token_ttl: int = 86400
    refresh_token_policy: RefreshTokenPolicyType = RefreshTokenPolicyType.static
    scope: set[str] = set()
    sector_identifier: str
    sector_identifier_uri: str | None = None
    token_endpoint_auth_method: ClientAuthenticationMethod

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        # When pairwise Subject Identifiers are used, the OpenID Provider
        # MUST calculate a unique sub (subject) value for each Sector
        # Identifier. The Subject Identifier value MUST NOT be reversible
        # by any party other than the OpenID Provider.
        #
        # Providers that use pairwise sub values and support Dynamic
        # Client Registration [OpenID.Registration] SHOULD use the
        # sector_identifier_uri parameter. It provides a way for a group
        # of websites under common administrative control to have consistent
        # pairwise sub values independent of the individual domain names.
        # It also provides a way for Clients to change redirect_uri
        # domains without having to reregister all of their users.
        #
        # If the Client has not provided a value for sector_identifier_uri
        # in Dynamic Client Registration [OpenID.Registration], the Sector
        # Identifier used for pairwise identifier calculation is the host
        # component of the registered redirect_uri. If there are multiple
        # hostnames in the registered redirect_uris, the Client MUST register
        # a sector_identifier_uri.
        #
        # When a sector_identifier_uri is provided, the host component of
        # that URL is used as the Sector Identifier for the pairwise identifier
        # calculation. The value of the sector_identifier_uri MUST be a URL
        # using the https scheme that points to a JSON file containing an
        # array of redirect_uri values. The values of the registered
        # redirect_uris MUST be included in the elements of the array.
        #
        # OpenId Connect Core 1.0, Section 8.1
        grant_types: list[GrantType] = values.get('grant_types') or []
        redirect_uris: list[str] = values.get('allowed_redirect_uris') or []
        sector_identifier: str | None = values.get('sector_identifier')
        sector_identifier_uri: str | None = values.get('sector_identifier_uri')
        redirect_hosts = set(
            filter(
                bool,
                [urllib.parse.urlparse(uri).hostname for uri in redirect_uris]
            )
        )

        needs_redirect_uri: bool = bool(set(grant_types) & {
            GrantType.authorization_code,
            GrantType.refresh_token
        })
        if len(redirect_uris) == 0 and needs_redirect_uri:
            raise ValueError("at least one redirect URI must be provided.")
        if not sector_identifier:
            if sector_identifier_uri is None:
                if len(redirect_hosts) > 1:
                    raise ValueError(
                        "sector_identifier or sector_identifier_uri must be "
                        "provided with multiple redirects URIs"
                    )
                sector_identifier = redirect_hosts.pop()
            else:
                assert isinstance(sector_identifier_uri, str)
                p = urllib.parse.urlparse(sector_identifier_uri)
                if not p.hostname:
                    raise ValueError('sector_identifier_uri does not contain a hostname.')
                sector_identifier = p.hostname
        values['sector_identifier'] = sector_identifier
        return values

    def can_grant(self, grant_type: GrantType) -> bool:
        return grant_type in self.grant_types

    def can_redirect(self, uri: RedirectURI) -> bool:
        """Return a boolean indicating if the client allows redirection
        to the given URI.
        """
        return uri in self.allowed_redirect_uris

    def can_use(self, scope: list[RequestedScope | str]) -> bool:
        return all([str(x) in self.scope for x in scope])

    def get_allowed_email_domains(self) -> set[DomainName]:
        if not self.downstream.providers: return set()
        return self.downstream.providers[0].allowed_email_domains or set()

    def get_contacts(self) -> list[str]:
        if not self.downstream.providers: return []
        return self.downstream.providers[0].contacts or []

    def get_pairwise_identifier(self, sub: int) -> PairwiseIdentifier:
        return PairwiseIdentifier(self.sector_identifier, sub)

    def get_redirect_uri(self, uri: RedirectURI | None) -> RedirectURI:
        if len(self.allowed_redirect_uris) > 1 and uri is None:
            raise FatalAuthorizationException(
                "The redirect_uri parameter is required because the client allows "
                "multiple redirect URIs."
            )
        uri = uri or self.allowed_redirect_uris[0]
        if uri not in self.allowed_redirect_uris:
            raise FatalAuthorizationException(
                "The client does not allow redirection to the given "
                "redirect_uri.",
                status_code=403
            )
        return RedirectURI(uri or self.allowed_redirect_uris[0])

    def requires_downstream(self) -> bool:
        return self.downstream.required

    def get_provider(self):
        assert self.downstream.providers
        if len(self.downstream.providers) > 1:
            raise NotImplementedError
        return self.downstream.providers[0]


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import cbra.core as cbra


class SymbolRequirement:
    __module__: str = 'cbra.core.ioc.test'


class InstanceRequirement:
    __module__: str = 'cbra.core.ioc.test'


class SingletonRequirement:
    __module__: str = 'cbra.core.ioc.test'


class NestedRequirement:
    __module__: str = 'cbra.core.ioc.test'

    def __init__(
        self,
        instance: InstanceRequirement = cbra.instance('InstanceRequirement'),
        singleton: SingletonRequirement = cbra.instance('SingletonRequirement')
    ) -> None:
        self.instance = instance
        self.singleton = singleton

    def __repr__(self) -> str:
        return f'NestedRequirement(instance={self.instance}, singleton={self.singleton})'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

from cbra.core.iam.models import Subject
from .basescope import BaseScope
from .iauthorizationrequest import IAuthorizationRequest
from .iresourceowner import IResourceOwner


class SubScope(BaseScope):
    name: Literal['sub']

    def apply(
        self,
        subject: Subject,
        owner: IResourceOwner,
        claims: dict[str, Any],
        request: IAuthorizationRequest | None = None
    ) -> None:
        claims['sub'] = str(owner.ppid.value)
        claims['sct'] = owner.ppid.sector


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cbra.types import ICache


class MemoryCache(ICache):
    __module__: str = 'cbra.ext.cache'
    objects: dict[str, Any] = {}

    def __init__(self) -> None:
        self.objects = MemoryCache.objects

    async def connect(self, config: Any) -> None:
        pass

    async def get(self, key: str) -> Any:
        return self.objects.get(key)
    
    async def set(self, key: str, value: Any, ttl: int | None = None) -> None:
        self.objects[key] = value


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
import urllib.request
from typing import NamedTuple

import fastapi
import fastapi.params

from .idependant import IDependant
from .malformedheader import MalformedHeader


ACCEPT_LANGUAGE_HEADER_MAX_LENGTH: int = 500


class RequestLanguage(IDependant):
    __module__: str = 'cbra.types'
    # Format of Accept-Language header values. From RFC 9110 Sections 12.4.2 and
    # 12.5.4, and RFC 5646 Section 2.1.
    pattern: re.Pattern[str] = re.compile(
        r"""
            # "en", "en-au", "x-y-z", "es-419", "*"
            ([A-Za-z]{1,8}(?:-[A-Za-z0-9]{1,8})*|\*)
            # Optional "q=1.00", "q=0.8"
            (?:\s*;\s*q=(0(?:\.[0-9]{,3})?|1(?:\.0{,3})?))?
            # Multiple accepts per header.
            (?:\s*,\s*|$)
        """,
        re.VERBOSE,
    )

    class AcceptedLanguage(NamedTuple):
        code: str
        weight: float


    @classmethod
    def parse_language(cls, language: str) -> AcceptedLanguage:
        if ';' not in language:
            return cls.AcceptedLanguage(language, 1.0)
        else:
            language, sep, weight = language.partition(';q=')
            if sep != ';q=': # pragma: no cover
                raise MalformedHeader
            try:
                return cls.AcceptedLanguage(language, float(weight))
            except (ValueError, TypeError): # pragma: no cover
                raise MalformedHeader

    def __init__(
        self,
        accept_language: str | None = fastapi.Header(
            default=None,
            alias='Accept-Language',
            title="Language(s)",
            description=(
                "The `Accept-Language` request HTTP header indicates the natural "
                "language and locale that the client prefers. The server uses "
                "content negotiation to select one of the proposals and informs "
                "the client of the choice with the `Content-Language` response "
                "header. Browsers set required values for this header according "
                "to their active user interface language."
            )
        )
    ) -> None:
        # If accept_language is a header instance, this was not invoked through
        # FastAPI dependency injection.
        if isinstance(accept_language, fastapi.params.Header):
            accept_language = None
        self.languages = []
        if accept_language is not None\
        and len(accept_language) <= ACCEPT_LANGUAGE_HEADER_MAX_LENGTH:
            if not self.pattern.match(accept_language):
                raise MalformedHeader
            self.languages = [
                self.parse_language(x)
                for x in urllib.request.parse_http_list(accept_language)
                if x != '*'
            ]
            self.languages = list(sorted(self.languages, key=lambda x: -x.weight))
        if not self.languages:
            self.languages.append(self.AcceptedLanguage('en', 1.0))

    def select(self, available: set[str], default: str = 'en') -> str:
        """Select a language based on the given available languages."""
        available = {str.lower(x) for x in available}
        for language in self.languages:
            if str.lower(language.code) not in available:
                continue
            break
        else:
            language = self.AcceptedLanguage(default, 1.0) 
        return language.code


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import logging
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Literal
from typing import TypeVar

import aorta
import fastapi

from .abortable import Abortable
from .iauthorizationcontextfactory import IAuthorizationContextFactory
from .irequestprincipal import IRequestPrincipal
from .iroutable import IRoutable
from .isessionmanager import ISessionManager
from .forbidden import Forbidden
from .notauthorized import NotAuthorized
from .request import Request
from .requestlanguage import RequestLanguage


T = TypeVar('T', bound='IEndpoint')


class IEndpoint:
    __module__: str = 'cbra.types'
    allowed_http_methods: list[str]
    autodiscover: bool = True
    context_factory: IAuthorizationContextFactory
    csrf_protect: bool = False
    handlers: list[IRoutable]
    include_in_schema: bool = True
    name: str | None = None
    languages: RequestLanguage = RequestLanguage.depends()
    logger: logging.Logger = logging.getLogger('cbra.endpoint')
    response_model_by_alias: bool = False
    response_model_exclude: set[str] | None = None
    response_model_exclude_none: bool = False
    session: ISessionManager[Any]
    timestamp: datetime
    publisher: aorta.types.IPublisher
    with_options: bool = True

    #: The ``If-Match`` value provided by the current request.
    etag: set[str] = set()

    #: The set of permissions supported by this endpoint. These must be
    #: defined beforehand to limit the number of calls to remote IAM
    #: systems.
    permissions: set[str]

    #: The list of subjects that may access this endpoint. Override this
    #: property if an endpoint needs to hardcode the allowed subjects
    #: that may invoke it, for example when expecting a request from
    #: Google using a priorly known service account.
    allowed_subjects: set[str] = set()

    #: The set of trusted authorization servers. This will override the
    #: :attr:`cbra.core.conf.settings.TRUSTED_AUTHORIZATION_SERVERS`
    #: setting.
    trusted_providers: set[str] = set()

    #: Indicates if this endpoint uses versioning using the ``ETag``
    #: and friends headers.
    versioned: bool

    principal: IRequestPrincipal
    request: Request
    response: fastapi.Response
    router: fastapi.APIRouter

    #: Indicates if all requests to the endpoint must be authenticated.
    require_authentication: bool = False
    status_code: int = 200
    summary: str | None = None
    tags: list[str] = []

    #: The current messaging transaction.
    transaction: aorta.Transaction

    #: Indicates that this endpoint is in testing mode and should not catch
    #: any abortable errors.
    test: bool = False

    @staticmethod
    def require_permission(name: str) -> Callable[..., Any]:
        """Decorate a method on an :class:`~cbra.types.IEndpoint` implementation
        to require the given permission. If the request does not have permission
        `name`, then :class:`~cbra.types.Forbidden` is raised.
        """
        def decorator_factory(
            func: Callable[..., Any]
        ) -> Callable[['IEndpoint'], Awaitable[Any]]:
            @functools.wraps(func)
            async def f(self: 'IEndpoint', *args: Any, **kwargs: Any) -> Any:
                if not await self.is_authorized(name):
                    raise Forbidden
                return await func(self, *args, **kwargs)
            return f
        return decorator_factory

    @classmethod
    def configure(cls: type[T], overrides: dict[str, Any]) -> type[T]:
        return type(cls.__name__, (cls,), overrides) # type: ignore

    async def get_allowed_subjects(self) -> set[str]:
        """Return the set of allowed subjects that may access this resource.
        By default, returns :attr:`allowed_subjects`.
        """
        return self.allowed_subjects

    def get_success_headers(self, data: Any) -> dict[str, Any]:
        """Return a mapping holding the headers to add on a successful
        request based on the return value of the request handler.
        """
        return {}

    async def run_handler(
        self,
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any
    ):
        try:
            async with aorta.Transaction(publisher=self.publisher) as tx:
                self.transaction = tx
                await self.authenticate()
                if self.must_authenticate() and not self.ctx.is_authenticated():
                    raise NotAuthorized
                return await func(self, *args, **kwargs)
        except Abortable as exc:
            return await exc.as_response()

    async def authenticate(self) -> None:
        self.ctx = await self.context_factory.authenticate(
            request=self.request,
            principal=self.principal,
            providers=self.trusted_providers,
            subjects=await self.get_allowed_subjects()
        )

    async def is_authorized(self, name: str) -> bool:
        """Return a boolean if the given authorization context has a
        certain permission.
        """
        await self.ctx.authorize()
        return self.has_permission(name)
    
    def has_permission(self, name: str) -> bool:
        """Return a boolean if the request has the given permission."""
        return self.ctx.has_permission(name)

    def is_authenticated(self) -> bool:
        """Return a boolean indicating if the request is authenticated."""
        return self.ctx.is_authenticated()

    def must_authenticate(self) -> bool:
        """Return a boolean if a client must authenticate the request."""
        return self.require_authentication

    def set_cookie(
        self,
        key: str,
        value: str = "",
        max_age: int | None = None,
        expires: datetime | str| int | None = None,
        path: str = "/",
        domain: str | None = None,
        secure: bool = False,
        httponly: bool = False,
        samesite: Literal["lax", "strict", "none"] | None = "lax",
    ) -> None:
        self.response.set_cookie(
            key=key,
            value=value,
            max_age=max_age,
            expires=expires,
            path=path,
            domain=domain,
            secure=secure,
            httponly=httponly,
            samesite=samesite
        )

    def delete_cookie(self, key: str, path: str = "/") -> None:
        return self.set_cookie(
            key,
            expires=datetime(1970, 1, 1, tzinfo=timezone.utc),
            max_age=0,
            path=path
        )
    
    def publish(
        self,
        Message: type[aorta.Command | aorta.Event],
        **params: Any
    ) -> None:
        self.transaction.publish(Message.parse_obj(params))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
from typing import Any
from typing import TypeVar

import pydantic
from google.cloud.datastore import Client as GoogleClient

import cbra.core as cbra
from cbra.ext.google.types import IDatastoreEntity
from cbra.ext.google.types import IDatastoreKey
from cbra.ext.google import BaseDatastoreRepository
from cbra.ext.security import ApplicationKeychain


T = TypeVar('T', bound=pydantic.BaseModel)


class BaseModelStorage(BaseDatastoreRepository):
    keychain: ApplicationKeychain

    def __init__(
        self,
        client: GoogleClient | Any = cbra.inject('GoogleDatastoreClient'),
        keychain: ApplicationKeychain = ApplicationKeychain.depends()
    ):
        if not isinstance(client, GoogleClient):
            raise TypeError(f"Invalid client: {repr(client)}")
        self.client = client
        self.keychain = keychain

    def entity_to_model(
        self,
        cls: type[T],
        entity: IDatastoreEntity | None,
        **kwargs: Any
    ) -> T | None:
        if entity is None:
            return None
        return cls.parse_obj({**dict(entity), **kwargs}) # type: ignore

    def model_to_entity(
        self,
        key: IDatastoreKey,
        obj: pydantic.BaseModel,
        exclude: set[str] | None = None,
        exclude_none: bool = False
    ) -> IDatastoreEntity:
        # TODO: this is only here to prevent encoding issues with protobuf
        entity = self.entity_factory(key)
        entity.update(json.loads(obj.json(exclude=exclude, exclude_none=exclude_none)))
        return entity


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
from typing import Any

import jinja2

from cbra.core.conf import settings
from cbra.core import translation


class BaseEmailSender:
    __module__: str = 'cbra.ext.email'

    @property
    def environment(self) -> jinja2.Environment:
        env = jinja2.Environment(
            loader=jinja2.ChoiceLoader([
                jinja2.FileSystemLoader('res/templates/email'),
                jinja2.PackageLoader('cbra.ext.email', 'templates'),
            ]),
            undefined=jinja2.StrictUndefined,
            extensions=[
                'jinja2.ext.i18n',
                'cbra.ext.email.AccessCodeExtension',
                'cbra.ext.email.ButtonExtension',
                'cbra.ext.email.ParagraphExtension',
                'cbra.ext.email.SectionExtension',
            ]
        )
        env.install_gettext_callables( # type: ignore
            gettext=translation.gettext,
            ngettext=translation.ngettext,
            newstyle=True
        )
        return env

    def get_template_context(self, ctx: dict[str, Any]) -> dict[str, Any]:
        """Update the given context with some default variables."""
        ctx = copy.deepcopy(ctx)
        ctx.update({
            'settings': settings
        })
        return ctx

    def get_template(self, template_name: str) -> jinja2.Template:
        """Return a :class:`jinja2.Template` instance using the
        configured environment (see :attr:`environment`).
        """
        return self.environment.get_template(template_name)
    
    def render_template(self, template_name: str, ctx: dict[str, Any]) -> str:
        """Render a template specified by `template_name` with the given
        context dictionary.
        """
        t = self.get_template(template_name)
        return t.render(**ctx)
    
    def render_templates(
        self,
        templates: dict[str, str],
        ctx: dict[str, Any]
    ) -> list[tuple[str, str]]:
        """Render the mapping of content types to template names."""
        return [
            (content_type, self.render_template(name, ctx))
            for content_type, name in templates.items()
        ]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import AsyncIterable

import fastapi

from cbra.types import Request
from ..connectionregistry import connections
from ..const import DEFAULT_CONNECTION_NAME
from ..session import Session


__all__: list[str] = ['DefaultSession']


async def get(request: Request) -> AsyncIterable[Session]:
    if connections.dispose not in request.app.teardown_handlers:
        request.app.teardown_handlers.add(connections.dispose)
    async with cast(Session, connections.session(DEFAULT_CONNECTION_NAME)) as session:
        yield session
    await session.close()

DefaultSession: Session = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import logging
from typing import Any

import fastapi
from headless.ext import oauth2
from headless.types import IClient
from headless.ext.oauth2.models import ManagedGrant
from headless.ext.oauth2.types import IAuthorizationServerClient
from headless.ext.oauth2.types import ITokenResponse

from cbra.core.conf import settings
from cbra.ext.credentials import CredentialRepository
from cbra.ext.credentials import ICredentialRepository
from cbra.ext.credentials.http import OpenAuthorizationAccessToken
from cbra.ext.credentials.types import SubjectCredentialIdentifier
from cbra.types import NotAuthorized
from cbra.types import NotFound
from ..types import IFrontendStorage
from ..types import ResourceServerAccessToken
from ..types import ResourceAccessTokenIdentifier
from .applicationclient import ApplicationClient
from .clientstorage import ClientStorage


logger: logging.Logger = logging.getLogger('cbra.auth')


async def cache_access_token(
    client: IAuthorizationServerClient,
    storage: IFrontendStorage,
    grant_id: str,
    audience: str,
    response: ITokenResponse
):
    access_token_id = ResourceAccessTokenIdentifier.parse_url(grant_id, audience)
    logger.info("Obtained access token (audience: %s)", access_token_id.resource)
    access_token = ResourceServerAccessToken.parse_response(
        grant_id=grant_id,
        issuer=client.get_issuer(),
        resource=access_token_id.resource,
        response=response # type: ignore
    )
    await storage.persist(access_token)


async def get(
    client: oauth2.Client = ApplicationClient,
    credentials: ICredentialRepository = CredentialRepository,
    storage: IFrontendStorage = ClientStorage,
    grant_id: SubjectCredentialIdentifier | None = fastapi.Cookie(
        default=None,
        title="Grant ID",
        alias='bff.grant',
        description=(
            "Identifies the grant that is used to exchange tokens with "
            "the authorization server."
        )
    ),
    resource: str = fastapi.Path(
        default=...
    ),
    path: str = fastapi.Path(
        default=...
    )
):
    path = path or '/'
    spec = settings.OAUTH2_RESOURCE_SERVERS.get(resource)
    if spec is None:
        raise NotFound
    service = spec['resource']
    if grant_id is None:
        logger.warning("Grant not specified (resource: %s)", service)
        raise NotAuthorized(
            headers={'WWW-Authenticate': 'authorization_code'}
        )
    grant = await credentials.get(grant_id)
    if grant is None:
        logger.warning("Grant does not exist (resource: %s).", service)
        raise NotAuthorized(
            headers={'WWW-Authenticate': 'authorization_code'}
        )
    if not isinstance(grant.credential, ManagedGrant):
        logger.critical("Invalid grant configured (id: %s)", grant_id)
        raise NotAuthorized(
            headers={'WWW-Authenticate': 'authorization_code'}
        )
    
    # Check if there is a cached access token for this resource
    # and grant.
    access_token_id = ResourceAccessTokenIdentifier.parse_url(
        grant_id=str(grant.get_primary_key()),
        service=service
    )
    cached = await storage.get(access_token_id)
    expires = token = None
    if cached is not None and not cached.is_expired():
        logger.debug("Using cached acccess token (service: %s)", service)
        expires = cached.expires
        token = cached.access_token
    assert isinstance(grant.credential, ManagedGrant)
    credential = OpenAuthorizationAccessToken(
        client=client,
        credential=grant,
        resource=service,
        token=token,
        expires=expires,
        on_obtained=functools.partial(cache_access_token, client, storage)
    )

    async with oauth2.ResourceServer(base_url=service, credential=credential) as resource_client:
        yield resource_client


RequestedResourceServerClient: IClient[Any, Any] = fastapi.Depends(get)




# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Test that the authorization endpoint properly authenticates the
subject.
"""
import urllib.parse

import pytest
from headless.core import httpx

from cbra.core.conf import settings
from cbra.ext import oauth2
from .conftest import SUPPORTED_RESPONSE_TYPES


@pytest.mark.parametrize("response_type", SUPPORTED_RESPONSE_TYPES)
@pytest.mark.asyncio
async def test_unauthenticated_redirects_to_login_endpoint(
    client: httpx.Client,
    authorization_endpoint: str,
    response_type: oauth2.types.ResponseType
):
    params: dict[str, str] = {
        'client_id': 'app',
        'redirect_uri': 'https://www.google.com',
        'response_type': response_type.value
    }
    response = await client.get(
        url=authorization_endpoint,
        params=params
    )
    assert response.status_code == 303, response.content
    assert 'Location' in response.headers
    p = urllib.parse.urlparse(response.headers['Location'])
    q = dict(urllib.parse.parse_qsl(p.query))
    assert 'next' in q, q
    assert 'request' in q, q


@pytest.mark.parametrize("response_type", SUPPORTED_RESPONSE_TYPES)
@pytest.mark.asyncio
async def test_unauthenticated_without_session_gets_session(
    client: httpx.Client,
    authorization_endpoint: str,
    response_type: oauth2.types.ResponseType
):
    params: dict[str, str] = {
        'client_id': 'app',
        'redirect_uri': 'https://www.google.com',
        'response_type': response_type.value
    }
    response = await client.get(
        url=authorization_endpoint,
        params=params
    )
    assert response.status_code == 303, response.content
    assert 'Set-Cookie' in response.headers
    assert settings.SESSION_COOKIE_NAME in response.impl.cookies
    p = urllib.parse.urlparse(response.headers['Location'])
    q = dict(urllib.parse.parse_qsl(p.query))
    assert 'next' in q
    assert 'request' in q


@pytest.mark.parametrize("response_type", SUPPORTED_RESPONSE_TYPES)
@pytest.mark.asyncio
async def test_evil_session_gets_new_session(
    evil_client: httpx.Client,
    authorization_endpoint: str,
    response_type: oauth2.types.ResponseType
):
    client = evil_client
    params: dict[str, str] = {
        'client_id': 'app',
        'redirect_uri': 'https://www.google.com',
        'response_type': response_type.value
    }
    response = await client.get(
        url=authorization_endpoint,
        params=params
    )
    assert response.status_code == 303, response.content
    assert 'Location' in response.headers
    p = urllib.parse.urlparse(response.headers['Location'])
    q = dict(urllib.parse.parse_qsl(p.query))
    assert 'next' in q
    assert 'request' in q

    assert 'Set-Cookie' in response.headers
    assert settings.SESSION_COOKIE_NAME in response.impl.cookies


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import timedelta
from typing import Any
from typing import Protocol


class IStorageBucket(Protocol):
    __module__: str = 'cbra.types'

    async def blob(self, path: str) -> Any:
        ...

    async def generate_signed_url(
        self,
        path: str,
        expires: timedelta | None = None,
        ttl: int | None = None,
        *args: Any,
        **kwargs: Any
    ) -> str:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone

import pydantic
from canonical import EmailAddress

from cbra.types import PersistedModel
from ..const import ALLOWED_DOMAINS
from .emailchallenge import EmailChallenge




class AccountRegistration(PersistedModel):
    """Retains the state of an account registration request that
    requires a backup email address.

    Registration of a new account is done by verifying an email address
    using a shared secret (verification code). If the email adres is not
    whitelisted per its domain, then a fallback email address must be
    provided. The :class:`AccountRegistration` model must then resolve
    either of three cases:

    - Both email addresses are not linked to existing accounts. In this
      case, an account that owns both addresses is created.
    - One address is linked to an existing account. The other address is
      then also added to the account.
    - The email addresses are linked to different accounts. This is an
      error state.

    Attributes:
      id (int): a numeric identifier for this :class:`AccountRegistration`.
      requested (datetime.datetime): the date/time at which registration
          of the account was requested.
    """
    id: int | None = pydantic.Field(
        default=None,
        auto_assign=True
    )

    requested: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )

    register: EmailChallenge = pydantic.Field(
        default=...,
        primary_key=True
    )

    fallback: EmailChallenge | None = None

    def add_fallback(self, email: EmailAddress) -> None:
        """Register a fallback email address."""
        if email.domain not in ALLOWED_DOMAINS:
            raise ValueError("Must be on a whitelisted domain.")
        self.fallback = EmailChallenge(email=email)

    def needs_backup(self) -> bool:
        """Return a boolean indicating if a backup email address must be
        provided to register an account.
        """
        return self.register.email.domain in ALLOWED_DOMAINS


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from headless.ext.oauth2 import Client
from headless.ext.oauth2.types import ResponseMode
from headless.ext.oauth2.types import ResponseType

import cbra.core as cbra
from ...params import ClientStorage
from ...types import IFrontendStorage
from ...types import AuthorizationState



class BaseAuthorizationEndpoint(cbra.Endpoint):
    """A :class:`~cbra.core.Endpoint` implementation for endpoints
    that initiate an OAuth 2.x/OpenID Connect authorization request.
    """
    __module__: str = 'cbra.ext.oauth2.client'
    callback_endpoint: str
    default_response_mode: str = ResponseMode.query.value
    storage: IFrontendStorage = ClientStorage

    def get_return_uri(self) -> str:
        """Return the URI to which the callback endpoint must redirect
        after succesfully obtaining the token from the authorization
        server.
        """
        return '/'

    def get_redirect_uri(self) -> str:
        """Return the URI to which the authorization server must redirect after
        the request is authorized.
        """
        return str(self.request.url_for(self.callback_endpoint))

    async def get(self) -> fastapi.Response:
        """Initiate an OAuth 2.x/OpenID Connect authorization request
        with default parameters provided by the server.
        """
        async with await self.get_client() as client:
            assert client.client_id is not None
            params = AuthorizationState.new(
                client_id=client.client_id,
                redirect_uri=self.get_redirect_uri(),
                response_mode=ResponseMode.query,
                response_type=ResponseType.code,
                return_uri=self.get_return_uri()
            )
            await self.storage.persist(params)
            return fastapi.responses.RedirectResponse(
                status_code=303,
                url=await params.get_authorize_uri(client)
            )

    async def get_client(self) -> Client:
        """Return a :class:`headless.ext.oauth2.Client` instance that can exchange
        the authorization code.
        """
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast

from cbra.types import IAuthorizationContext
from cbra.types import ISubject
from cbra.types import Request
from cbra.types import PolicyPrincipal
from ..types import RFC9068AccessToken
from .types import UserInfoSubject


class ResourceServerAuthorizationContext(IAuthorizationContext):
    __module__: str = 'cbra.ext.oauth2.resource'
    _request: Request
    _subject: UserInfoSubject
    _token: RFC9068AccessToken

    def __init__(
        self,
        request: Request,
        token: RFC9068AccessToken,
        subject: ISubject
    ):
        self._request = request
        self._subject = cast(UserInfoSubject, subject)
        self._token = token

    def get_principals(self) -> set[PolicyPrincipal]:
        principals: set[PolicyPrincipal] = set()
        if self.subject.email is not None:
            principals.add(PolicyPrincipal.email(self.subject.email))

        # Only add the client_id as a principal if it is the same
        # as the subject.
        if self._token.client_id == self._token.sub:
            principals.add(PolicyPrincipal.client(self._token.iss, self._token.client_id))

        return principals

    def get_subject(self) -> ISubject:
        return self._subject


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# type: ignore
from typing import Any
from typing import Generic
from typing import TypeVar

from .icredential import ICredential


P = TypeVar('P')


class ICredentialVerifier(Generic[P]):
    """Knows how to verify a credential attached to a principal."""
    __module__: str = 'cbra.types'

    async def verify(
        self,
        principal: P,
        credential: ICredential | None,
        providers: set[str] | None = None
    ) -> bool:
        raise NotImplementedError


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .icredential import ICredential
from .verifier import Verifier


class HMACSignature(ICredential):
    __module__: str = 'cbra.types'

    def __init__(self, signature: bytes, message: bytes):
        self.message = message
        self.signature = signature

    async def verify(self, verifier: Verifier) -> bool:
        return await verifier.verify(self.signature, self.message)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

from .compositeobjectidentifier import CompositeObjectIdentifier
from .iaccesstoken import IAccessToken


class ResourceAccessTokenIdentifier(CompositeObjectIdentifier[IAccessToken]):
    __module__: str = 'cbra.ext.oauth2.types'

    @classmethod
    def parse_url(cls, grant_id: str, service: str):
        p = urllib.parse.urlparse(service)
        return cls(grant_id, p.netloc)

    def __init__(self, grant_id: str, resource: str):
        self.grant_id = grant_id
        self.resource = resource

    def __repr__(self) -> str:
        return f"<ResourceAccessTokenIdentifier('{self.grant_id}', '{self.resource}')>"


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections
import random
from typing import cast
from typing import Any
from typing import TypeVar

from cbra.core.iam import MemorySubjectRepository
from cbra.core.iam.models import Subject
from .basestorage import BaseStorage
from .models import AuthorizationRequestParameters
from .models import AuthorizationState
from .models import AuthorizationServerModel
from .models import Client
from .models import ManagedGrant
from .models import RefreshToken
from .models import ResourceOwner
from .types import AuthorizationCode
from .types import AuthorizationRequestIdentifier
from .types import PairwiseIdentifier


T = TypeVar('T')


class MemoryStorage(BaseStorage):
    """An in-memory implementation of :class:`~cbra.ext.oauth2.BaseStorage`."""
    __module__: str = 'cbra.ext.oauth2'
    objects: dict[str, dict[Any, Any]] = collections.defaultdict(dict)
    subjects: MemorySubjectRepository = MemorySubjectRepository()

    @classmethod
    def clear(cls) -> None:
        cls.objects = collections.defaultdict(dict)

    def __init__(self):
        self.objects = MemoryStorage.objects

    async def destroy(self, obj: AuthorizationServerModel) -> None:
        self.objects[type(obj).__name__].pop(obj.__key__[0][1], None) # type: ignore

    async def get_authorization_request_by_id(
        self,
        oid: AuthorizationRequestIdentifier
    ) -> AuthorizationRequestParameters | None:
        return self._get(AuthorizationRequestParameters, oid)
    
    async def get_authorization_request_by_code(
        self,
        oid: AuthorizationCode
    ) -> AuthorizationRequestParameters | None:
        for params in self.objects[AuthorizationRequestParameters.__name__].values():
            if not isinstance(params, AuthorizationRequestParameters):
                continue
            if params.code is None or params.code.value != oid:
                continue
            break
        else:
            params = None
        return params

    async def get_resource_owner(self, id: int) -> ResourceOwner | None:
        return self._get(ResourceOwner, id)

    async def get_state(self, key: str) -> AuthorizationState | None:
        return self._get(AuthorizationState, key)

    async def persist_authorization_request(self, obj: AuthorizationRequestParameters) -> None:
        assert obj.id is not None
        self.objects[type(obj).__name__][obj.id] = obj

    async def persist_client(self, obj: Client) -> None:
        self.objects[type(obj).__name__][obj.client_id] = obj

    async def persist_managed_grant(
        self,
        obj: ManagedGrant
    ) -> None:
        self.objects[type(obj).__name__][obj.id] = obj

    async def persist_ppid(self, obj: PairwiseIdentifier) -> None:
        obj.value = random.randint(10**9, (10**10) - 1)
        self.objects[type(obj).__name__][obj.value] = obj

    async def persist_refresh_token(self, obj: RefreshToken) -> None:
        self.objects[type(obj).__name__][obj.token] = obj

    async def persist_resource_owner(self, obj: ResourceOwner) -> None:
        self.objects[type(obj).__name__][obj.id] = obj

    async def persist_state(self, obj: AuthorizationState) -> None:
        self.objects[type(obj).__name__][obj.state] = obj

    async def persist_subject(self, obj: Subject) -> None:
        await self.subjects.persist(obj) # type: ignore

    def _get(self, Model: type[T], key: int | str) -> T | None:
        return cast(Model, self.objects[Model.__name__].get(key))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .tokenendpointexception import TokenEndpointException


class InvalidTarget(TokenEndpointException):
    error: str = 'invalid_target'
    message: str = (
        "The requested resource is invalid, missing, unknown, or malformed."
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic


class PostgreSQLParameters(pydantic.BaseModel):
    engine: Literal['postgresql'] = pydantic.Field(
        alias='ENGINE'
    )

    host: str = pydantic.Field(
        alias='HOST'
    )

    port: int = pydantic.Field(
        alias='PORT'
    )

    name: str = pydantic.Field(
        alias='NAME'
    )

    user: str = pydantic.Field(
        alias='USER'
    )

    password: str = pydantic.Field(
        alias='PASSWORD'
    )

    @property
    def dsn(self) -> str:
        return f'postgresql+asyncpg://{self.user}:{self.password}@{self.host}:{self.port}/{self.name}'


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio

import pytest

from cbra.core.translation import LanguageContext


@pytest.mark.asyncio
async def test_activate():
    async def f():
        LanguageContext.activate('foo')
    assert LanguageContext.get() == 'en'
    await f()
    assert LanguageContext.get() == 'foo'


@pytest.mark.asyncio
async def test_activate_nested():
    LanguageContext.activate('foo')
    assert LanguageContext.get() == 'foo'
    with LanguageContext('en'):
        assert LanguageContext.get() == 'en'
    assert LanguageContext.get() == 'foo'


@pytest.mark.asyncio
async def test_nested_translations():
    assert LanguageContext.get() == 'en'

    async def f1():
        assert LanguageContext.get() == 'nl'
        with LanguageContext('en'):
            assert LanguageContext.get() == 'en'
        assert LanguageContext.get() == 'nl'

    with LanguageContext('nl'):
        assert LanguageContext.get() == 'nl'
        await f1()
        assert LanguageContext.get() == 'nl'


@pytest.mark.asyncio
async def test_concurrent():
    assert LanguageContext.get() == 'en'

    async def f1():
        with LanguageContext('foo'):
            await asyncio.sleep(1)
            assert LanguageContext.get() == 'foo'


    async def f2():
        async def f2_n1():
            assert LanguageContext.get() == 'bar'
            with LanguageContext('baz'):
                assert LanguageContext.get() == 'baz'
            assert LanguageContext.get() == 'bar'

        await asyncio.sleep(0.5)
        with LanguageContext('bar'):
            assert LanguageContext.get() == 'bar'
            await f2_n1()

    async def f3():
        await asyncio.sleep(0.25)
        assert LanguageContext.get() == 'en'

    await asyncio.gather(f1(), f2(), f3())


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable

import fastapi

from .apiroute import APIRoute
from .endpointprovider import EndpointProvider
from .utils import parent_signature


class APIRouter(fastapi.APIRouter, EndpointProvider):
    __module__: str = 'cbra.core'

    @parent_signature(fastapi.APIRouter.__init__)
    def __init__(self, *args: Any, **kwargs: Any):
        kwargs.setdefault('route_class', APIRoute)
        super().__init__(*args, **kwargs)

    @parent_signature(fastapi.APIRouter.add_api_route)
    def add_api_route(
        self,
        endpoint: Callable[..., Any],
        *args: Any,
        **kwargs: Any
    ) -> None:
        self.update_requirements(endpoint)
        return super().add_api_route(endpoint=endpoint, *args, **kwargs)

    def add_to_router(
        self,
        router: Any,
        *args: Any,
        **kwargs: Any
    ) -> None:
        router.include_router(self, *args, **kwargs)

    @parent_signature(fastapi.APIRouter.head)
    def head(self, *a: Any, **k: Any):
        return self.discover_requirements(fastapi.APIRouter.head, *a, **k)

    @parent_signature(fastapi.APIRouter.get)
    def get(self, *a: Any, **k: Any):
        return self.discover_requirements(fastapi.APIRouter.get, *a, **k)

    @parent_signature(fastapi.APIRouter.post)
    def post(self, *a: Any, **k: Any):
        return self.discover_requirements(fastapi.APIRouter.post, *a, **k)

    @parent_signature(fastapi.APIRouter.patch)
    def patch(self, *a: Any, **k: Any):
        return self.discover_requirements(fastapi.APIRouter.patch, *a, **k)

    @parent_signature(fastapi.APIRouter.put)
    def put(self, *a: Any, **k: Any):
        return self.discover_requirements(fastapi.APIRouter.put, *a, **k)

    @parent_signature(fastapi.APIRouter.trace)
    def trace(self, *a: Any, **k: Any):
        return self.discover_requirements(fastapi.APIRouter.trace, *a, **k)

    @parent_signature(fastapi.APIRouter.options)
    def options(self, *a: Any, **k: Any):
        return self.discover_requirements(fastapi.APIRouter.options, *a, **k)

    @parent_signature(fastapi.APIRouter.delete)
    def delete(self, *a: Any, **k: Any):
        return self.discover_requirements(fastapi.APIRouter.delete, *a, **k)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

import pydantic
from canonical import ResourceName
from headless.ext.oauth2.models import ManagedGrant

from cbra.types import PersistedModel
from cbra.ext.security import ApplicationKeychain
from .types import CredentialScope
from .types import GlobalCredentialIdentifier
from .types import ICredentialRepository


class GlobalCredential(PersistedModel):
    scope: Literal[CredentialScope.GLOBAL] = pydantic.Field(
        default=CredentialScope.GLOBAL
    )

    service: ResourceName = pydantic.Field(
        default=...,
        primary_key=True
    )

    credential: ManagedGrant = pydantic.Field(
        default=...
    )

    encrypted: bool = pydantic.Field(
        default=False
    )

    __repository__: ICredentialRepository | None = pydantic.PrivateAttr(
        default=None
    )

    def get_primary_key(self) -> GlobalCredentialIdentifier:
        return GlobalCredentialIdentifier(self.service)

    async def decrypt(self, keychain: ApplicationKeychain):
        await self.credential.decrypt(keychain)
        self.encrypted = False

    async def encrypt(self, keychain: ApplicationKeychain):
        await self.credential.encrypt(keychain)
        self.encrypted = True

    async def persist(self) -> None:
        assert self.__repository__ is not None
        await self.__repository__.persist(self)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from datetime import datetime
from datetime import timezone
from typing import TypeVar

import pydantic
from cbra.types import PersistedModel
from headless.ext.oauth2 import Client
from headless.ext.oauth2.types import ResponseMode
from headless.ext.oauth2.types import ResponseType


T = TypeVar('T', bound='AuthorizationState')


class AuthorizationState(PersistedModel):
    client_id: str = pydantic.Field(
        default=...
    )

    created: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    nonce: str = pydantic.Field(
        default=...
    )

    redirect_uri: str = pydantic.Field(
        default=...
    )

    return_uri: str = pydantic.Field(
        default='/'
    )

    response_mode: ResponseMode = pydantic.Field(
        default=...
    )

    response_type: ResponseType = pydantic.Field(
        default=...
    )

    state: str = pydantic.Field(
        default=...,
        primary_key=True
    )

    scope: set[str] = pydantic.Field(
        default_factory=set
    )

    @classmethod
    def new(
        cls: type[T],
        *,
        client_id: str,
        redirect_uri: str,
        response_mode: ResponseMode,
        response_type: ResponseType,
        return_uri: str = '/'
    ) -> T:
        return cls(
            client_id=client_id,
            nonce=secrets.token_urlsafe(48),
            redirect_uri=redirect_uri,
            return_uri=return_uri,
            response_mode=response_mode,
            response_type=response_type,
            state=secrets.token_urlsafe(48),
        )
    
    def is_valid(
        self,
        state: str
    ) -> bool:
        """Return a boolean indicating if the authorization server returned
        a valid response.
        """
        return all([
            self.state == state
        ])
    
    def is_openid(self) -> bool:
        return 'openid' in self.scope

    async def get_authorize_uri(self, client: Client) -> str:
        params: dict[str, str] = {}
        if self.is_openid():
            params['nonce'] = self.nonce
        return await client.authorize(
            state=self.state,
            redirect_uri=self.redirect_uri
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timezone
import logging
from typing import Literal

from headless.ext.oauth2.models import TokenResponse
from headless.ext.oauth2.types import GrantType

from ..models import Client
from ..types import IAuthorizationServerStorage
from ..types import ITokenBuilder
from ..types import ITokenSigner
from ..types import InvalidScope
from ..types import InvalidTarget
from ..types import RequestedScope
from .basetokenrequest import BaseTokenRequest


logger: logging.Logger = logging.getLogger('cbra.endpoint')


class ClientCredentialsGrant(BaseTokenRequest):
    client: Client
    grant_type: Literal[GrantType.client_credentials]
    scope: set[str] = set()
    resource: str

    async def create_response(
        self,
        storage: IAuthorizationServerStorage,
        builder: ITokenBuilder,
        signer: ITokenSigner
    ) -> TokenResponse:
        if not self.client.can_use(list(self.scope)):
            raise InvalidScope
        if not self.client.allows_resources({self.resource}):
            raise InvalidTarget
        at, expires_in = builder.rfc9068(
            sub=self.client.client_id,
            scope=[RequestedScope.parse_obj({'name': x}) for x in self.scope],
            auth_time=int(datetime.now(timezone.utc).timestamp()),
            audience=self.resource
        )
        access_token = await at.sign(signer)
        await self.register_issued_access_token(
            token=at,
            signed=access_token,
            scope=list(self.scope),
            storage=storage,
            sub=self.client.client_id,
            claims={
                'sub': self.client.client_id
            }
        )
        return TokenResponse(
            access_token=access_token, # type: ignore
            expires_in=expires_in,
            token_type='Bearer',
            scope=str.join(' ', sorted(self.scope))
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from google.cloud.datastore import Client

from .environ import GOOGLE_DATASTORE_NAMESPACE
from .environ import GOOGLE_SERVICE_PROJECT


class DefaultDatastoreClient(Client):

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__( # type: ignore
            project=GOOGLE_SERVICE_PROJECT,
            namespace=GOOGLE_DATASTORE_NAMESPACE,
            *args,
            **kwargs
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import ipaddress
import urllib.parse
from typing import Any
from typing import Callable
from typing import Generator
from typing import TypeVar

from pydantic.validators import str_validator

T = TypeVar('T', bound='RedirectURI')


OOB_URLS: set[str] = {
    "urn:ietf:wg:oauth:2.0:oob",
    "urn:ietf:wg:oauth:2.0:oob:auto"
}


class RedirectURI(str):
    __module__: str = 'cbra.ext.oauth2.types'
    params: dict[str, str]

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: dict[str, Any]
    ) -> None:
        field_schema.update( # pragma: no cover
            title='Email Address',
            type='string',
            format='email'
        )

    @classmethod
    def __get_validators__(cls: type[T]) -> Generator[Callable[..., str | T], None, None]:
        yield str_validator
        yield cls.validate

    @classmethod
    def validate(cls: type[T], v: str) -> T:
        if len(v) > 2048:
            raise ValueError("value is too long to be a valid URL.")
        if v in OOB_URLS:
            raise ValueError('OOB is a security risk.')

        p = urllib.parse.urlparse(v)
        if not p.scheme or not p.netloc:
            raise ValueError('not a valid URL.')
        if p.query:
            raise ValueError('query parameters must not be used.')
        if p.fragment:
            raise ValueError('fragments must not be used.')

        # Check if the value is an IP address.
        if p.hostname:
            try:
                ip = ipaddress.ip_address(p.hostname)
            except ValueError:
                # Not a valid ip address. The hostname points to a domain
                if p.scheme != 'https':
                    raise ValueError('the https scheme must be used.')
            else:
                if not ip.is_loopback and p.scheme != 'https':
                    raise ValueError('the https scheme must be used.')

            if p.hostname == 'localhost':
                # Clients should use loopback IP literals rather than the
                # string localhost as described in Section 8.4.2.
                # (OAuth 2.1 draft).
                raise ValueError('local redirect URIs must use loopback IP literals.')

        return cls(urllib.parse.urlunparse(p))

    def redirect(self, allow_params: bool = False, **params: Any) -> str:
        """Create a redirect URI with the given params."""
        p: list[str] = list(urllib.parse.urlparse(self)) # type: ignore
        params = {k: v for k, v in params.items() if v is not None}
        if allow_params:
            params.update(urllib.parse.parse_qsl(p[4]))
        p[4] = urllib.parse.urlencode(params, quote_via=urllib.parse.quote)
        return urllib.parse.urlunparse(p)

    def __repr__(self) -> str: # pragma: no cover
        return f'RedirectURI({self})'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

from .resourcemodel import ResourceModel

T = TypeVar('T', bound=ResourceModel)


class QueryResult(Generic[T]):
    __module__: str = 'cbra.core'
    token: str | None

    def __init__(self, token: str | None, items: list[T]):
        self.token = token or None
        self.items = items


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeAlias
from typing import TypeVar
from typing import Union

from headless.ext.oauth2 import OIDCToken

from .models import ManagedGrant
from .types import AuthorizationState
from .types import AuthorizationStateIdentifier
from .types import IFrontendStorage
from .types import ManagedGrantIdentifier
from .types import CompositeObjectIdentifier
from .types import ObjectIdentifier
from .types import OIDCTokenSubjectIdentifier
from .types import ResourceAccessTokenIdentifier
from .types import ResourceServerAccessToken


FrontendModel: TypeAlias = Union[
    AuthorizationState,
    ManagedGrant,
    OIDCToken,
    ResourceServerAccessToken
]

Identifier: TypeAlias = OIDCTokenSubjectIdentifier

T = TypeVar('T')


class BaseFrontendStorage(IFrontendStorage):
    __module__: str = 'cbra.ext.oauth2'
    supported_types: tuple[type, ...] = (
        AuthorizationState,
        ManagedGrant,
        OIDCToken,
        ResourceServerAccessToken,
    )

    async def get(self, oid: ObjectIdentifier[T] | CompositeObjectIdentifier[T]) -> None | T:
        if isinstance(oid, OIDCTokenSubjectIdentifier):
            return await self.get_oidc_token(oid) # type: ignore
        elif isinstance(oid, ManagedGrantIdentifier):
            return await self.get_grant(oid) # type: ignore
        elif isinstance(oid, ResourceAccessTokenIdentifier):
            return await self.get_access_token(oid) # type: ignore
        elif isinstance(oid, AuthorizationStateIdentifier):
            return await self.get_authorization_state(oid) # type: ignore

    async def get_access_token(
        self,
        oid: ResourceAccessTokenIdentifier
    ) -> ResourceServerAccessToken | None:
        raise NotImplementedError

    async def get_authorization_state(self, oid: AuthorizationStateIdentifier) -> AuthorizationState | None:
        raise NotImplementedError

    async def get_grant(self, oid: ManagedGrantIdentifier) -> ManagedGrant | None:
        raise NotImplementedError

    async def get_oidc_token(self, oid: OIDCTokenSubjectIdentifier) -> OIDCToken | None:
        raise NotImplementedError

    async def persist(self, obj: FrontendModel) -> None:
        if not isinstance(obj, self.supported_types):
            raise TypeError(type(obj).__name__)
        if isinstance(obj, AuthorizationState):
            await self.persist_authorization_state(obj)
        if isinstance(obj, ResourceServerAccessToken):
            await self.persist_access_token(obj)
        if isinstance(obj, ManagedGrant):
            await self.persist_grant(obj)
        if isinstance(obj, OIDCToken):
            await self.persist_oidc_token(obj)

    async def persist_access_token(self, obj: ResourceServerAccessToken) -> None:
        """Persist a :class:`~cbra.ext.oauth2.ResourceServerAccessToken`
        instance.
        """
        raise NotImplementedError

    async def persist_authorization_state(self, obj: AuthorizationState) -> None:
        """Persist a :class:`~cbra.ext.oauth2.AuthorizationState` instance."""
        raise NotImplementedError

    async def persist_grant(self, obj: ManagedGrant) -> None:
        """Persist a :class:`~cbra.ext.oauth2.models.ManagedGrant`
        instance.
        """
        raise NotImplementedError
    
    async def persist_oidc_token(self, obj: OIDCToken) -> None:
        """Persist a :class:`~headless.ext.oauth2.OIDCToken` instance."""
        raise NotImplementedError
        



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Awaitable
from typing import Callable

import pydantic

from cbra.core.conf import settings
from cbra.core.ioc.loader import import_symbol
from cbra.types import ICache


class CacheConfiguration(pydantic.BaseModel):
    factory: Callable[[], Awaitable[ICache] | ICache]
    namespace: str = ''
    qualname: str

    @classmethod
    def get(cls, name: str) -> 'CacheConfiguration':
        spec = settings.CACHES.get(name)
        if spec is None:
            raise ValueError(f"No such cache: {name}")
        return cls.parse_obj(spec)
    
    @pydantic.root_validator(pre=True) # type: ignore
    def preprocess(
        cls,
        values: dict[str, str]
    ) -> dict[str, str]:
        qualname = values.get('qualname')
        if qualname is not None:
            values['factory'] = import_symbol(qualname)
        return values


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Awaitable
from typing import Callable
from typing import TypeVar

import fastapi
from canonical import EmailAddress

import cbra.core as cbra
from cbra.types import IDependant
from ..types import ClientInfo
from ..types import FatalAuthorizationException
from ..types import IAuthorizationServerStorage
from ..types import OIDCProvider
from ..types import PairwiseIdentifier
from ..types import RedirectURI
from ..types import RequestedScope
from .client import Client


T = TypeVar('T', bound='AuthorizationRequestClient')


class AuthorizationRequestClient(IDependant):
    """An OAuth 2.x client that is requesting authorization."""
    __module__: str = 'cbra.ext.oauth2.models'
    __client: Client
    client_id: str

    @classmethod
    async def fromrequest(
        cls: type[T],
        client_id: str | None = fastapi.Query(
            default=None,
            title="Client ID",
            description="Identifies the client that is requesting authorization."
        ),
        storage: IAuthorizationServerStorage = cbra.ioc.instance('_AuthorizationServerStorage')
    ) -> T:
        if client_id is None:
            raise FatalAuthorizationException("The client_id parameter is required.")
        client = await storage.get(Client, client_id)
        if client is None:
            raise FatalAuthorizationException("The client does not exist.")
        return cls(client)

    @property
    def client_info(self) -> ClientInfo:
        return self.__client.client_info

    @property
    def impl(self) -> Client:
        return self.__client

    @property
    def sector_identifier(self) -> str:
        return self.__client.sector_identifier

    @classmethod
    def __inject__(cls: type[T]) -> Callable[..., Awaitable[T] | T]:
        return cls.fromrequest

    def __init__(self, client: Client):
        self.__client = client # type: ignore
        self.client_id = client.client_id

    def allows_email(self, email: EmailAddress) -> bool:
        return True

    def allows_resources(self, resources: set[str]) -> bool:
        return self.__client.allows_resources(resources)

    def can_use(self, scope: list[RequestedScope]) -> bool:
        return self.__client.can_use(scope)

    def can_redirect(self, uri: RedirectURI) -> bool:
        return self.__client.can_redirect(uri)

    def get_redirect_uri(self, uri: RedirectURI | None) -> RedirectURI:
        return self.__client.get_redirect_uri(uri)

    def get_pairwise_identifier(self, sub: int) -> PairwiseIdentifier:
        return self.__client.get_pairwise_identifier(sub)

    def has_acl(self) -> bool:
        return self.__client.has_acl()

    def is_member(self, obj: Any) -> bool:
        return self.__client.is_member(obj) if obj else False

    def requires_downstream(self) -> bool:
        return self.__client.requires_downstream()
    
    def get_provider(self) -> OIDCProvider:
        return self.__client.get_provider()


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`LocalAortaTransport`."""
import asyncio
import base64
import datetime
import logging
import os
from typing import Any
from typing import Awaitable

from aorta.types import Envelope
from aorta.types import ITransport
from httpx import AsyncClient

import cbra.core as cbra
from .messagepublished import MessagePublished
from .pubsubmessage import PubsubMessage


class LocalAortaTransport(ITransport): # pragma: no cover
    """A :class:`aorta.transport.ITransport` implementation that sends
    messages to the ASGI application provided to the constructor.
    """
    logger: logging.Logger = logging.getLogger('cbra.endpoint')

    def __init__(self, app: cbra.Application, base_url: str, path: str):
        self.app = app
        self.base_url = base_url
        self.path = path

    async def send(self, messages: list[Envelope[Any]], is_retry: bool = False):
        """Construct a :class:`~cbra.ext.google.models.MessagePublished`
        instance and send it to the Eventarc endpoint.
        """
        futures: list[Awaitable[None]] = []
        async with AsyncClient(app=self.app, base_url=self.base_url) as client:
            for obj in messages:
                message = MessagePublished(
                    subscription="projects/abcdef/subscriptions/abcdef",
                    message=PubsubMessage(
                        messageId=bytes.hex(os.urandom(16)),
                        publishTime=datetime.datetime.utcnow(),
                        attributes={},
                        data=bytes.decode(base64.b64encode(bytes(obj)))
                    )
                )
                futures.append(self._send(client, message))

            await asyncio.gather(*futures)

    async def _send(self, client: AsyncClient, message: MessagePublished):
        response = await client.post( # type: ignore
            url=self.path,
            data=message.json(by_alias=True), # type: ignore
            headers={
                'Accept': "application/json",
                'Content-Type': "application/json"
            }
        )
        if response.status_code not in {200, 202}:
            self.logger.error(
                "Caught non-200 response from Aorta endpoint: %s",
                response.status_code
            )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import TypeVar

from pydantic.fields import ModelField

from cbra.types import ModelInspector
from cbra.types import PersistedModel
from .types import IDatastoreEntity


T = TypeVar('T', bound=PersistedModel)
D = TypeVar('D')


class PolymorphicDatatoreCursor:
    inspector: ModelInspector = ModelInspector()
    items: list[PersistedModel] = []
    model: type[PersistedModel]
    primary_key: ModelField
    token: str | None

    def __init__(
        self,
        model: type[PersistedModel],
        entities: list[IDatastoreEntity],
        token: bytes | str | None = None,
        adapter: Callable[[PersistedModel], Any] | None = None
    ):
        if isinstance(token, bytes):
            token = bytes.decode(token, 'utf-8')
        self.model = model
        self.primary_key = self.inspector.get_primary_key_field(model) # type: ignore
        self.token = token
        assert self.primary_key is not None
        self.adapter = adapter or (lambda e: e)
        self.items = [self.restore(e) for e in entities]

    def restore(self, entity: IDatastoreEntity) -> Any:
        obj = self.model.parse_obj({
            **dict(entity),
            self.primary_key.name: entity.key.id or entity.key.name
        })
        return self.adapter(obj)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from ckms.core import Keychain
from ckms.core import KeySpecification
from ckms.jose import PayloadCodec

from cbra.types import IDependant
from ..const import DEFAULT_SIGNING_KEY
from .serverkeychain import ServerKeychain




class TokenSigner(IDependant):
    __module__: str = 'cbra.ext.oauth2.params'
    keychain: Keychain
    signing_key: KeySpecification

    @property
    def alg(self) -> str:
        return str(self.signing_key.algorithm)
    
    @property
    def crv(self) -> str | None:
        return self.signing_key.curve

    def __init__(self, keychain: Keychain = ServerKeychain):
        self.codec = PayloadCodec(
            signer=keychain,
            signing_keys=[DEFAULT_SIGNING_KEY]
        )
        self.keychain = keychain
        self.signing_key = keychain.get(DEFAULT_SIGNING_KEY)
        assert self.signing_key is not None

    async def jwt(
        self,
        claims: dict[str, Any],
        typ: str = 'jwt'
    ) -> str:
        return await self.codec.encode(
            payload=claims,
            content_type=typ
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from headless.ext.oauth2 import OIDCToken

from cbra.core.iam import RequestSubject
from cbra.types import IRequestPrincipal


class OIDCRequestSubject(RequestSubject):
    authenticated: bool = False
    token: OIDCToken
    principal: IRequestPrincipal
    sub: str

    def __init__(
        self,
        *,
        id: str,
        token: OIDCToken,
        principal: IRequestPrincipal
    ):
        self.email = token.email
        self.principal = principal
        self.sub = id
        self.token = token

    def get_display_name(self) -> str:
        name = self.token.name
        if not self.token.name:
            if self.token.given_name and self.token.family_name:
                # TODO: not compatible with east Asian names.
                name = f'{self.token.given_name} ${self.token.family_name}'
            elif self.token.given_name or self.token.family_name:
                name = self.token.given_name or self.token.family_name
            else:
                name = self.token.email
        assert name is not None
        return name


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
import secrets
from datetime import datetime
from datetime import timezone

from ckms.jose import Decoder
from ckms.types import Malformed

from .bearertokenexception import BearerTokenException
from .itokensigner import ITokenSigner
from .requestedscope import RequestedScope
from .signable import Signable


class RFC9068AccessToken(Signable):
    iss: str
    aud: str| list[str]
    exp: int
    sub: str
    client_id: str
    iat: int
    jti: str
    auth_time: int | None = None
    acr: str = '0'
    amr: list[str] | None = []
    scope: str | None = None

    @classmethod
    def new(
        cls,
        client_id: str,
        iss: str,
        aud: str | set[str] | list[str],
        sub: str,
        now: int,
        ttl: int,
        scope: list[RequestedScope] | None = None,
        auth_time: int | None = None,
        acr: str = '0',
        amr: list[str] | None = None
    ) -> 'RFC9068AccessToken':
        if isinstance(aud, (list, set)):
            aud = list(sorted(aud))
        
        params: dict[str, int | str | list[str] | None] = {
            'acr': acr,
            'aud': aud,
            'client_id': client_id,
            'exp': now + ttl,
            'iat': now,
            'iss': iss,
            'jti': secrets.token_urlsafe(48),
            'nbf': now,
            'sub': sub,
            'auth_time': auth_time
        }
        if scope is not None:
            params['scope'] = ' '.join(sorted([x.name for x in scope]))
        if amr is not None:
            params['amr'] = amr
        return cls.parse_obj(params)

    @classmethod
    def parse_jwt(cls, token: str) -> 'RFC9068AccessToken':
        typ = None
        try:
            jose, jwt = Decoder.introspect(token)
            if jwt is None:
                raise Malformed
            for header in jose.headers:
                if header.typ is None:
                    continue
                typ = str.lower(header.typ)
                break
            else:
                typ = None
            if typ is None or typ != "at+jwt":
                raise Malformed
        except Malformed:
            raise
            raise BearerTokenException(
                scheme='Bearer',
                error='invalid_token',
                error_description=(
                    "The access token provided is expired, revoked, malformed, "
                    "or invalid for other reasons."
                )
            )
        return cls.parse_obj(jwt.dict())

    def get_expire_date(self) -> datetime:
        """Return an aware :class:`datetime.datetime` instance
        representing the date and time that the access token
        expires.
        """
        return datetime.fromtimestamp(float(self.exp), timezone.utc)

    def get_scope(self) -> set[str]:
        return set(filter(bool, re.split(r'\s+', self.scope or '')))

    def is_expired(self) -> bool:
        return self.get_expire_date() < datetime.now(timezone.utc)

    def verify_audience(self, allow: set[str]):
        audience = set([self.aud] if isinstance(self.aud, str) else self.aud)
        if not (audience & allow):
            raise BearerTokenException(
                scheme='Bearer',
                error='invalid_token',
                error_description=(
                    "The audience is not accepted by the resource server."
                )
            )

    def verify_scope(self, require: set[str]):
        scope = set(filter(bool, re.split(r'\s+', self.scope or '')))
        if scope < require:
            raise BearerTokenException(
                scheme='Bearer',
                error='insufficient_scope',
                error_description=(
                    "The access token was not granted the required scope for this "
                    "resource."
                ),
                status_code=403
            )

    def verify_expired(self):
        if self.is_expired():
            raise BearerTokenException(
                scheme='Bearer',
                error='invalid_token',
                error_description="The access token is expired."
            )

    def verify_issuers(self, allow: set[str]):
        if self.iss not in allow:
            raise BearerTokenException(
                scheme='Bearer',
                error='invalid_token',
                error_description=(
                    "The access token was issued by an untrusted authorization server."
                )
            )

    def verify(
        self,
        audiences: set[str],
        issuers: set[str],
        scope: set[str] | None = None
    ):
        self.verify_audience(audiences)
        self.verify_expired()
        self.verify_issuers(issuers)
        self.verify_scope(scope or set())

    async def sign(self, signer: ITokenSigner) -> str:
        return await signer.jwt(self.dict(exclude_none=True), "at+jwt")


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from cbra.core.translation import utils


@pytest.mark.parametrize("language,result", [
    ("en_US", "en-us"),
    ("en", "en"),
])
def test_locale_to_language(language: str,  result: str):
    assert utils.to_language(language) == result


@pytest.mark.parametrize("language,result", [
    ("en-us", "en_US"),
    ("sr-latn", "sr_Latn"),
    ("en", "en"),
])
def test_language_to_locale(language: str,  result: str):
    assert utils.to_locale(language) == result



# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
import pytest

import cbra.core as cbra


def dependency():
    return 1

def f(value: int = fastapi.Depends(dependency)):
    return value


@pytest.mark.asyncio
async def test_run_with_injected_value():
    result = await cbra.run(f)
    assert result == 1


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.ext.email import EmailChallengeEndpoint


class EmailRegistrationEndpoint(EmailChallengeEndpoint):
    __module__: str = 'cbra.ext.oauth2.server'
    name: str = 'oauth2.register.email'
    path_name: str = 'email'
    path: str = '/register/email'
    summary: str = 'Email Registration Endpoint'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Iterable


class Scope:
    """Represents the scope required to access an endpoint or resource."""
    __module__: str = 'cbra.ext.oauth2.resource.types'
    requires: set[str]

    def __init__(self, scope: str | Iterable[str]):
        if isinstance(scope, str):
            scope = [scope]
        self.requires = set(scope)

    def missing(self, granted: set[str]) -> set[str]:
        """Return the set containing the scope that is missing."""
        return self.requires - granted


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from headless.ext.oauth2 import Client
from headless.ext.oauth2 import TokenResponse
from headless.ext.oauth2.models import AuthorizationCode

from cbra.types import IDependant
from .authorizationstateidentifier import AuthorizationStateIdentifier
from .authorizationstate import AuthorizationState



class AuthorizeCodeResponse(pydantic.BaseModel, IDependant):
    code: str = pydantic.Field(
        default=...
    )

    iss: str | None = pydantic.Field(
        default=None
    )

    state: AuthorizationStateIdentifier = pydantic.Field(
        default=...
    )

    async def obtain(self, client: Client, params: AuthorizationState) -> TokenResponse:
        return await client.exchange_authorization_code(
            dto=AuthorizationCode(code=self.code, state=params.state),
            redirect_uri=str(params.redirect_uri)
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
import datetime

import aorta
import pytest
import pytest_asyncio
from headless.core import httpx

import cbra.core as cbra
from cbra.ext.google import AortaEndpoint
from cbra.ext.google import MessagePublished



AortaEndpoint: type[AortaEndpoint] = AortaEndpoint.configure( # type: ignore
    {'require_authentication': False}
)


@pytest.fixture
def app() -> cbra.Application:
    return cbra.Application()


@pytest_asyncio.fixture # type: ignore
async def client(app: cbra.Application):
    async with httpx.Client(base_url='http://cbra.ext.google', app=app) as client:
        yield client


@pytest.mark.asyncio
async def test_accept_message(
    app: cbra.Application,
    client: httpx.Client,
):
    app.add(AortaEndpoint)

    event = GoogleEvent(foo=1)
    envelope = event.envelope()
    message = MessagePublished.parse_obj({
        'subscription': 'projects/myproject/subscriptions/mysub',
        'message': {
            'messageId': 1,
            'publishTime': datetime.datetime.now().isoformat(),
            'data': base64.b64encode(bytes(envelope))
        }
    })
    response = await client.post(
        url='/',
        content=message.json(by_alias=True)
    )
    assert response.status_code == 202, response.content
    raise Exception

class GoogleEvent(aorta.Event):
    foo: int


class GoogleEventListener(aorta.EventListener):
    dep = cbra.inject('Missing')

    async def handle(self, event: GoogleEvent) -> int:
        return event.foo
    

aorta.register(GoogleEventListener)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import urllib.parse
import warnings
from typing import Any

from canonical import EmailAddress
from canonical import StringType


class PolicyPrincipal(StringType):
    """Identifies a member (principal) in an access control list or
    role membership. A :class:`PolicyPrincipal` may be one of the following
    types:

    - `domain` - a specific domain. Domain membership is established by
      proving control over an email address for the domain.
    - `user` - a specific user, identified by email address.
    - `sub` - a specific subject, identified by subject identifier.
    """
    __module__: str = 'cbra.types'
    principal_types: set[str] = {
        'client',
        'domain',
        'user',
        'sub',
    }

    @classmethod
    def client(cls, issuer: str, client_id: str):
        p = urllib.parse.urlparse(issuer)
        if p.path.lstrip('/') != '':
            raise NotImplementedError("Can not use issuer with path component.")
        return cls(f'client:{client_id}@{p.netloc}')

    @classmethod
    def email(cls, email: str | EmailAddress):
        """Create a new :class:`PolicyPrincipal` of type ``user``
        given an email address.
        """
        return cls(f'user:{email}')

    @classmethod
    def validate(cls, v: str) -> str:
        v = super().validate(v)
        t, _ = str.split(v, ':', 1)
        if t not in cls.principal_types:
            raise ValueError(f'invalid principal type: {t}')
        return cls(v)
    
    @functools.singledispatchmethod
    def match(self, obj: Any) -> bool:
        warnings.warn(
            f"Unknown princpal for access policy: {type(obj).__name__}"
        )
        return False
    
    @match.register
    def match_email(self, email: EmailAddress) -> bool:
        kind, value = str.split(self, ':', 1)
        if kind not in {'domain', 'user', 'sub'}:
            return False
        is_matching = False
        if kind == 'domain':
            is_matching = value == str(email.domain)
        elif kind == 'user':
            is_matching = value == str(email)
        return is_matching



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

import pytest
from headless.core import httpx


@pytest.mark.parametrize("state", [
    None,
    'state'
])
@pytest.mark.asyncio
async def test_missing_response_type_returns_303_error(
    client: httpx.Client,
    authorization_endpoint: str,
    state: str | None,
    issuer: str
):
    params: dict[str, str] = {'client_id': 'app', 'response_type': 'foo'}
    if state is not None:
        params['state'] = state
    response = await client.get(
        url=authorization_endpoint,
        params=params
    )
    assert response.status_code == 303, response.content
    assert 'Location' in response.headers
    p = urllib.parse.urlparse(response.headers['Location'])
    q = dict(urllib.parse.parse_qsl(p.query))
    assert 'error' in q
    assert q['error'] == "invalid_request"
    assert 'iss' in q
    assert q['iss'] == issuer
    if state is not None:
        assert 'state' in q
        assert q['state'] == state


@pytest.mark.parametrize("state", [
    None,
    'state'
])
@pytest.mark.asyncio
async def test_invalid_response_type_returns_303_error(
    client: httpx.Client,
    authorization_endpoint: str,
    state: str | None,
    issuer: str
):
    params: dict[str, str] = {'client_id': 'app', 'response_type': 'foo'}
    if state is not None:
        params['state'] = state
    response = await client.get(
        url=authorization_endpoint,
        params=params
    )
    assert response.status_code == 303, response.content
    assert 'Location' in response.headers
    p = urllib.parse.urlparse(response.headers['Location'])
    q = dict(urllib.parse.parse_qsl(p.query))
    assert 'error' in q
    assert 'iss' in q
    assert q['iss'] == issuer
    assert q['error'] == "invalid_request"
    if state is not None:
        assert 'state' in q
        assert q['state'] == state


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any

import pydantic
import pytest

from cbra.ext.oauth2.types import RedirectURI


@pytest.mark.parametrize("value", [
    "https://google.com/callback",
    "http://127.0.0.1/callback",
    "http://127.0.0.1:8000/callback",
    "http://[::1]/callback",
    "http://[::1]:8000/callback",
    "https://google.com/",
    "http://127.0.0.1/",
    "http://127.0.0.1:8000/",
    "http://[::1]/",
    "http://[::1]:8000/",
])
def test_valid_values(value: Any):
    for validate in RedirectURI.__get_validators__():
        validate(value)


@pytest.mark.parametrize("value", [
    "invalid",
    None,
    'a' * 2049,
    'http://rp.example.com/callback',
    'http://localhost:8000/callback',
    'https://localhost:8000/callback',
    'https://localhost:8000/callback?foo=bar',
    'https://localhost:8000/callback#fragment',
    'http://1.1.1.1/callback',
    "urn:ietf:wg:oauth:2.0:oob",
    "urn:ietf:wg:oauth:2.0:oob:auto",
])
def test_invalid_values(value: Any):
    with pytest.raises((pydantic.errors.PydanticTypeError, ValueError)):
        for validate in RedirectURI.__get_validators__():
            validate(value)


def test_cast_to_string():
    url = RedirectURI('https://rp.example.com')
    assert str(url) == 'https://rp.example.com'


def test_create_redirect_uri():
    url = RedirectURI(urllib.parse.urlparse('https://rp.example.com'))
    p = urllib.parse.urlparse(url.redirect(foo='bar'))
    q = dict(urllib.parse.parse_qsl(p.query))
    assert 'foo' in q
    assert q.get('foo') == 'bar'


def test_serialize_model():
    class M(pydantic.BaseModel):
        redirect_uri: RedirectURI

        class Config:
            json_encoders: dict[type, Any] = {
                RedirectURI: str
            }

    obj = M.parse_obj({
        'redirect_uri': 'https://rp.example.com/callback'
    })
    obj.json()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.core import Resource

from .frontendauthenticationcontextfactory import FrontendAuthorizationContextFactory


class FrontendResource(Resource, model=None):
    __abstract__: bool = True
    __module__: str = 'cbra.ext.bff'
    autodiscover: bool = False
    context_factory: FrontendAuthorizationContextFactory
    require_authentication: bool = True


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import importlib
import logging
from typing import Any

from sqlalchemy.ext.asyncio import async_sessionmaker
from sqlalchemy.ext.asyncio import create_async_engine
from sqlalchemy.ext.asyncio import AsyncEngine
from sqlalchemy.ext.asyncio import AsyncSession

from .types import ConnectionParameters


class ConnectionRegistry:
    __module__: str = 'cbra.ext.sql'
    connections: dict[str, ConnectionParameters] = {}
    logger: logging.Logger = logging.getLogger('cbra')
    settings: Any = None
    _engines: dict[str, AsyncEngine] = {}

    def __init__(self):
        self._engines = ConnectionRegistry._engines

    def dsn(self, name: str) -> str:
        return self.params(name).dsn

    async def dispose(self) -> None:
        """Close all engines."""
        self.logger.info("Disposing RDBMS connections")
        for name, engine in self._engines.items():
            self.logger.debug("Disposing %s", name)
            await engine.dispose()

    def session(self, name: str, cls: type[AsyncSession] = AsyncSession) -> AsyncSession:
        factory: async_sessionmaker[cls] = async_sessionmaker(self.get(name), expire_on_commit=False)
        return factory()

    def params(self, name: str) -> ConnectionParameters:
        if not self.settings:
            self.settings = importlib.import_module('cbra.core.conf').settings
            for connection_name, params in self.settings.DATABASES.items():
                self.connections[connection_name] = ConnectionParameters.parse_obj(params)
        return self.connections[name]       

    def get(self, name: str) -> AsyncEngine:
        """Return an :class:`sqlalchemy.ext.asyncio.AsyncEngine` instance."""
        params = self.params(name)
        if name not in self._engines:
            self._engines[name] = create_async_engine(
                params.dsn,
                pool_size=25,
                pool_timeout=120.0
            )
        return self._engines[name]
    

connections: ConnectionRegistry = ConnectionRegistry()

del ConnectionRegistry


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable
from typing import Coroutine

import fastapi
import fastapi.exceptions
import fastapi.routing

from cbra.types import Request


class APIRoute(fastapi.routing.APIRoute):

    def get_route_handler(
        self
    ) -> Callable[[fastapi.Request], Coroutine[Any, Any, fastapi.Response]]:
        super_handler = super().get_route_handler()

        async def handler(request: fastapi.Request) -> fastapi.Response:
            return await super_handler(
                Request(request.scope, request.receive)
            )

        return handler


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from headless.ext.oauth2.models import OIDCToken
from headless.ext.oauth2.models import TokenResponse

from cbra.ext.credentials import CredentialRepository
from cbra.ext.credentials import SubjectCredential
from cbra.ext.credentials import ICredentialRepository
from cbra.types import SessionRequestPrincipal
from cbra.core.conf import settings
from .client import ManagedGrant
from .params import FrontendError
from .params import FrontendTokenResponse
from .tokenhandlerendpoint import TokenHandlerEndpoint


class FrontendRedirectionEndpoint(TokenHandlerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    credentials: ICredentialRepository = CredentialRepository
    error: Exception | None = FrontendError
    name: str = 'bff.redirection'
    path: str = '/oauth/v2/callback'
    principal: SessionRequestPrincipal # type: ignore
    status_code: int = 303
    redirect_uri: str = fastapi.Cookie(
        default=...,
        title="Redirect URI",
        alias='bff.redirect_uri',
        description=(
            "The URI to redirect the user-agent to after completing the "
            "login."
        )
    )
    summary: str = 'Frontend Redirection Endpoint'
    token: TokenResponse = FrontendTokenResponse

    async def get(self) -> fastapi.Response:
        if self.error:
            raise self.error
        if self.token.refresh_token is None:
            raise NotImplementedError
        if self.token.id_token is None:
            raise NotImplementedError
        assert self.client.client_id is not None
        oidc = oidc = OIDCToken.parse_jwt(self.token.id_token)
        if oidc.email:
            oidc.email_verified = settings.APP_ISSUER_TRUST
        credential = SubjectCredential(
            subject_id=oidc.sub,
            resource=self.request.url.netloc,
            credential=ManagedGrant(
                client_id=self.client.client_id,
                iss=self.client.get_issuer(),
                oidc=oidc,
                refresh_token=self.token.refresh_token,
                scope=set() if not self.token.scope else set(filter(bool, self.token.scope.split(' '))),
                token_endpoint=None
            )
        )

        await self.session
        await self.credentials.persist(credential)
        self.session.authenticate(oidc)
        self.delete_cookies()
        self.set_cookie(
            'grant', credential.id,
            httponly=True,
            secure=True,
            samesite='strict'
        )
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=self.redirect_uri
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic
from canonical import EmailAddress
from ckms.types import JSONWebKeySet
from headless.ext.oauth2.types import GrantType

from cbra.ext import secrets
from ..types import ClientAuthenticationMethod
from ..types import ClientInfo
from ..types import OIDCProvider
from ..types import PairwiseIdentifier
from ..types import RedirectURI
from ..types import RefreshTokenPolicyType
from ..types import RequestedScope
from .applicationclient import ApplicationClient


class Client(pydantic.BaseModel):
    __root__: ApplicationClient

    @property
    def access_token_ttl(self) -> int | None:
        return None

    @property
    def auth_method(self) -> ClientAuthenticationMethod:
        return self.__root__.token_endpoint_auth_method

    @property
    def client_id(self) -> str:
        return self.__root__.client_id

    @property
    def client_info(self) -> ClientInfo:
        return ClientInfo(
            allowed_email_domains=self.__root__.get_allowed_email_domains(),
            contacts=self.__root__.get_contacts(),
            id=self.__root__.client_id,
            name=self.__root__.client_name,
            organization_name=self.__root__.organization_name,
            sector_identifier=self.__root__.sector_identifier
        )

    @property
    def jwks(self) -> JSONWebKeySet | None:
        return self.__root__.jwks

    @property
    def refresh_token_policy(self) -> RefreshTokenPolicyType:
        return self.__root__.refresh_token_policy

    @property
    def refresh_token_ttl(self) -> int:
        return self.__root__.refresh_token_ttl

    @property
    def sector_identifier(self) -> str:
        return self.__root__.sector_identifier

    def allows_email(self, email: EmailAddress) -> bool:
        return True

    def allows_resources(self, resources: set[str]) -> bool:
        if self.__root__.allowed_resources is None: return True
        return resources <= set(self.__root__.allowed_resources)

    def can_grant(self, grant_type: GrantType) -> bool:
        return self.__root__.can_grant(grant_type)

    def can_redirect(self, uri: RedirectURI) -> bool:
        return self.__root__.can_redirect(uri)

    def can_use(self, scope: list[RequestedScope | str]) -> bool:
        return self.__root__.can_use(scope)

    def get_redirect_uri(self, uri: RedirectURI | None) -> RedirectURI:
        return self.__root__.get_redirect_uri(uri)
    
    def get_pairwise_identifier(self, sub: int) -> PairwiseIdentifier:
        return self.__root__.get_pairwise_identifier(sub)

    def requires_downstream(self) -> bool:
        return self.__root__.requires_downstream()
    
    def get_provider(self) -> OIDCProvider:
        return self.__root__.get_provider()

    def has_acl(self) -> bool:
        """Return a boolean indicating if the client has an ACL."""
        return self.__root__.acl.is_enabled()

    def is_confidential(self) -> bool:
        return self.auth_method != ClientAuthenticationMethod.none

    def is_member(self, obj: Any) -> bool:
        """Return a boolean indicating if the object is part of the
        ACL.
        """
        return self.__root__.acl.is_member(obj)

    async def get_client_secret(self) -> str | None:
        secret = self.__root__.client_secret
        if isinstance(secret, dict):
            container = await secrets.load(secrets.parse(**secret))
            secret = container.get_value()
        return secret